
import numpy as np
from napari.layers import Image
from skimage import io

from cellpose_counter.counter import (
    estimate_diameter,
    get_image_patches,
    get_image_roi,
    restore_image,
)

SMALL_IMAGE_PATH = "src/cellpose_counter/_tests/data/image_small_512.tif"
LARGE_IMAGE_PATH = "src/cellpose_counter/_tests/data/image_large_1024.tif"

SMALL_IMAGE = io.imread(SMALL_IMAGE_PATH, as_gray=True)
LARGE_IMAGE = io.imread(LARGE_IMAGE_PATH, as_gray=True)

SMALL_ROI_LAYER = [np.array([[0.0, 256.0], [0.0, 512.0], [256.0, 512.0], [256.0, 256.0]])]
LARGE_ROI_LAYER = [np.array([[0.0, 0.0], [0.0, 1024.0], [1024.0, 1024.0], [1024.0, 0.0]])]

# Targets
SMALL_IMAGE_DIAMETER = np.float64(7.0)
SMALL_ROI_SHAPE = (
    (SMALL_ROI_LAYER[0][1][1] - SMALL_ROI_LAYER[0][0][1]).astype(int),
    (SMALL_ROI_LAYER[0][2][1] - SMALL_ROI_LAYER[0][3][1]).astype(int),
)

def test_roi_extraction():
    roi, _ = get_image_roi(SMALL_IMAGE, SMALL_ROI_LAYER[0])
    assert roi.shape == SMALL_ROI_SHAPE

def test_estimate_diameter():
    model_type = "nuclei"
    use_gpu = False # for testing
    diam = estimate_diameter(SMALL_IMAGE, model_type, use_gpu)
    assert diam == SMALL_IMAGE_DIAMETER

def test_image_restoration():
    restore_type = "nuclei_oneclick"
    use_gpu = False
    diameter = 6.48
    batch_size = 8

    image_layer = Image(SMALL_IMAGE, name="small_image_512")
    restored_images = restore_image(image_layer, None, restore_type, diameter, use_gpu, batch_size)

    assert len(restored_images) == 1
    assert restored_images[0].data.shape == (512, 512)

def test_image_split() -> None:
    """
    Test batch processing is enabled for larger images
    """
    target_dim = (512, 512)
    split_image_scale_threshold = (1.5, 1.5)

    small_batch = get_image_patches(SMALL_IMAGE, target_dim, split_image_scale_threshold)
    n_rows_small = np.ceil(SMALL_IMAGE.shape[0] / target_dim[0])
    n_cols_small = np.ceil(SMALL_IMAGE.shape[1] / target_dim[1])
    n_small_image_patches = n_rows_small * n_cols_small

    large_batch = get_image_patches(
        LARGE_IMAGE, target_dim, split_image_scale_threshold
    )
    n_rows_large = np.ceil(LARGE_IMAGE.shape[0] / target_dim[0])
    n_cols_large = np.ceil(LARGE_IMAGE.shape[1] / target_dim[1])
    n_large_image_patches = n_rows_large * n_cols_large

    assert len(list(small_batch)) == n_small_image_patches
    assert len(list(large_batch)) == n_large_image_patches




from typing import TYPE_CHECKING

from magicgui.widgets import (
    CheckBox,
    Combobox,
    Container,
    FloatSlider,
    LineEdit,
    PushButton,
    create_widget,
)
from napari.qt.threading import create_worker
from napari.utils import notifications

if TYPE_CHECKING:
    import napari
    import numpy as np


CP_MODELS = ["nuclei", "cyto3", "cyto2", "cyto"]
RESTORE_MODELS = [
    "denoise_nuclei",
    "deblur_nuclei",
    "upsample_nuclei",
    "oneclick_nuclei",
    "denoise_cyto3",
    "deblur_cyto3",
    "upsample_cyto3",
    "oneclick_cyto3",
    "denoise_cyto2",
    "deblur_cyto2",
    "upsample_cyto2",
    "oneclick_cyto2",
    "None"
]


class Counter(Container):
    """
    Counter widget for Napari

    count cells and nuclei from images using cellpose models.
    """
    def __init__(self, viewer: "napari.viewer.Viewer"):
        super().__init__()
        self._viewer = viewer
        # add a default ROI layer
        self._viewer.add_shapes(
            name="ROI",
            face_color="white",
            edge_color="red",
            edge_width=3,
            opacity=0.20,
            blending="additive",
        )
        # make scale bar visible
        self._viewer.scale_bar.visible = True
        self._viewer.scale_bar.unit = "px"

        # create widgets
        self._image_layer_combo = create_widget(
            label="Input Image", annotation="napari.layers.Image"
        )
        self._roi_layer_combo = create_widget(
            label="ROI", annotation="napari.layers.Shapes"
        )
        self._use_gpu = CheckBox(text="Use GPU", value=True)
        self._batch_size = LineEdit(label="Initial Batch Size", value="64")

        # image restoration
        self._restore_models = Combobox(label="Restore Model", value="oneclick_nuclei", choices=RESTORE_MODELS)
        self._restore_image_btn = PushButton(label="Restore image")

        # diameter estimation
        self._diam = LineEdit(label="Diameter", value="17")
        self._diam_estimate_btn = PushButton(label="Estimate Diameter")

        # segmentation
        self._cp_models = Combobox(label="Cellpose Model", value="nuclei", choices=CP_MODELS)
        self._cellprob_threshold = FloatSlider(label="Probability Threshold", value=0, min=-8.0, max=8.0, step=0.2)
        self._flow_threshold = FloatSlider(label="Flow Threshold", value=0.4, min=0.0, max=3.0, step=0.05)
        self._get_count_btn = PushButton(text="Get Automated Count")
        self._auto_count_display = LineEdit(label="Auto Count", value="0")
        self._manual_count_display = LineEdit(label="Manual Count", value="0")
        self._total_count_display = LineEdit(label="Total Count", value="0")

        # manual count adjustments
        self._auto_count_layer = create_widget(label="Auto Counts", annotation="napari.layers.Labels")
        self._manual_count_layer = create_widget(label="Manual Counts", annotation="napari.layers.Points")
        self._sync_count_btn = PushButton(text="Sync Auto+Manual Counts")

        # callbacks
        self._diam_estimate_btn.changed.connect(self._estimate_diam)
        self._restore_image_btn.changed.connect(self._restore_image)
        self._get_count_btn.changed.connect(self._get_count)
        self._sync_count_btn.changed.connect(self._update_count)

        # append into/extend the container with widgets
        self.extend([
            self._image_layer_combo,
            self._roi_layer_combo,
            self._use_gpu,
            self._batch_size,
            self._restore_models,
            self._restore_image_btn,
            self._diam,
            self._diam_estimate_btn,
            self._cp_models,
            self._cellprob_threshold,
            self._flow_threshold,
            self._get_count_btn,
            self._auto_count_display,
            self._manual_count_display,
            self._total_count_display,
            self._auto_count_layer,
            self._manual_count_layer,
            self._sync_count_btn
        ])


    def _display_diam(self, diam: float):
        """Purely for updating the diameter value"""
        self._diam.value = diam
        notifications.show_info("Successfully calculated diameter.")


    def _display_restored_image(self, image: "napari.layers.Image"):
        """Display restored image in the viewer"""
        self._viewer.add_layer(image)
        notifications.show_info("Successfully restored image.")


    def _display_count(self, result: tuple["np.ndarray", tuple[int, int], float]):
        """Display count and add segmentation results"""
        masks, mask_offsets, count = result
        self._total_count_display.value = count
        self._manual_count_display.value = 0
        self._auto_count_display.value = count

        self._viewer.add_labels(masks, translate=mask_offsets, name="Masks")
        self._viewer.add_points(name="Manual", size=float(self._diam.value))
        self._manual_count_layer.value = self._viewer.layers[-1]
        self._auto_count_layer.value = self._viewer.layers[-2]
        notifications.show_info("Updated total count.")


    def _estimate_diam(self):
        """
        Estimate diameter of objects with Cellpose size models

        Will use the entire image if no ROI selection is made. Otherwise,
        will use the first drawn ROI.
        """

        notifications.show_info("Calculating diameter...")
        image_layer = self._image_layer_combo.value
        if image_layer is None:
            notifications.show_error("No image selected.")
            return

        roi_layer = self._roi_layer_combo.value
        if roi_layer is None or len(roi_layer.data) == 0:
            image = image_layer.data
        else:
            from .counter import get_image_roi
            if len(roi_layer.data) > 1:
                notifications.show_warning("Only 1 ROI per Shape layer is allowed. Defaulting to 1st ROI.")
            image, _ = get_image_roi(image_layer.data, roi_layer.data[0])

        from .counter import estimate_diameter
        estimate_diam_worker = create_worker(estimate_diameter, image, self._cp_models.value, self._use_gpu.value, _progress=True)
        estimate_diam_worker.returned.connect(self._display_diam)
        estimate_diam_worker.start()


    def _restore_image(self):
        notifications.show_info("Restoring image...")
        image_layer = self._image_layer_combo.value
        if image_layer is None:
            notifications.show_error("No image selected.")
            return

        roi_layer = self._roi_layer_combo.value
        if roi_layer is None or len(roi_layer.data) == 0:
            roi = None
        else:
            if len(roi_layer.data) > 1:
                notifications.show_warning("Only 1 ROI per Shape layer allowed. Defaulting to 1st ROI.")
            roi = roi_layer.data[0]

        from .counter import restore_image
        restore_image_worker = create_worker(
            restore_image,
            image_layer,
            roi,
            self._restore_models.value,
            float(self._diam.value),
            self._use_gpu.value,
            int(self._batch_size.value),
            _progress=True
        )
        restore_image_worker.returned.connect(self._display_restored_image)
        restore_image_worker.start()


    def _get_count(self):
        """Count cells/nuclei inside an ROI for a given image using cellpose"""
        notifications.show_info("Counting objects...")
        image = self._image_layer_combo.value
        roi = self._roi_layer_combo.value

        if roi is not None and len(roi.data) > 1:
            notifications.show_warning("Only 1 ROI per shape layer allowed. Defaulting to 1st ROI.")

        if image is None:
            notifications.show_error("No image found in viewer.")
            return

        from .counter import count_objects
        count_objects_worker = create_worker(
            count_objects,
            image,
            roi,
            self._cp_models.value,
            self._cellprob_threshold.value, # cellprob and flow thresholds are
            self._flow_threshold.value,     # already floats. No need to convert.
            self._use_gpu.value,
            int(self._batch_size.value),
            float(self._diam.value),
            _progress=True
        )
        count_objects_worker.returned.connect(self._display_count)
        count_objects_worker.start()


    def _update_count(self):
        """Update total number of cells from automatic and manual counting"""
        count = 0
        manual_layer = self._manual_count_layer.value
        auto_layer = self._auto_count_layer.value

        if manual_layer is not None:
            manual_count = len(manual_layer.data)
            count += manual_count
            self._manual_count_display.value = manual_count

        if auto_layer is not None:
            auto_count = auto_layer.data.max()
            count += auto_count
            self._auto_count_display.value = auto_count

        self._total_count_display.value = count
        notifications.show_info("Total count updated successfully.")




import itertools
from typing import TYPE_CHECKING, Any, Iterator

import numpy as np
from accelerate import find_executable_batch_size
from napari.layers import Image, Shapes
from napari.utils import notifications
from skimage.transform import resize

if TYPE_CHECKING:
    pass


def get_image_roi(image: np.ndarray, roi_coords: np.ndarray) -> tuple[np.ndarray, tuple[int, int]]:
    """
    Get ROI from image

    :param image: Image data (np.ndarray)
    :param rois: List of ROI coordinates (list[np.ndarray])

    :returns:
        Iterator of tuples containing image ROIs (np.ndarray) and ROI offsets (tuple[int, int])
    """


    minr, minc = np.min(roi_coords, axis=0).astype(int)
    maxr, maxc = np.max(roi_coords, axis=0).astype(int)

    return image[minr:maxr, minc:maxc], (minr, minc)


def estimate_diameter(image: np.ndarray, model_type: str, use_gpu: bool) -> np.ndarray[Any, np.dtype[np.float16]]:
    """
    Estimate diameters using cellpose size models.

    :param image: Image data (np.ndarray)
    :param model_type: Cellpose model type (str)
    :param use_gpu: whether to use gpu acceleration (bool)

    :returns:
        Estimated diameter (np.ndarray[Any, np.dtype[np.float16]])
    """

    from cellpose import models
    model = models.Cellpose(model_type=model_type, gpu=use_gpu)
    diameter = model.sz.eval(image, channels=[0,0], channel_axis=-1)[0]
    diameter = np.round(diameter, 2)

    del model
    notifications.show_info(f"Estimated Diameter (px): {diameter}")
    return diameter


def restore_image(image_layer: Image, roi: np.ndarray | None, restore_type: str, diameter: float, use_gpu: bool, batch_size: int) -> Image:
    """
    Run Cellpose image restoration

    :param image_layer: Image layer (napar.layers.Image).
    :param restore_type: Restoration type/model (str).
    :param diameter: Object diameter in image (float).
    :param use_gpu: Whether to use GPU acceleration (bool).
    :param batch_size: Batch size for model inference (int).

    :returns:
        Restored image napari.layers.Image
    """

    restore_name = restore_type.split("_")[0]
    image_name = f"{image_layer.name}_{restore_name}"
    from cellpose.denoise import DenoiseModel
    restore_model = DenoiseModel(model_type=restore_type, diam_mean=diameter, gpu=use_gpu)

    if roi is None:
        restored_image = restore_model.eval([image_layer.data], channels=[0,0], channel_axis=-1, batch_size=batch_size)
    else:
        image, _ = get_image_roi(image_layer.data, roi)
        restored_image = restore_model.eval([image], channels=[0,0], channel_axis=-1, batch_size=batch_size)

    del restore_model
    return Image(restored_image[0][:,:,0], name=image_name, rgb=False)


def get_image_patches(
    image: np.ndarray,
    image_patch_dim: tuple[int, int],
    split_image_threshold: tuple[float, float],
) -> Iterator[tuple[np.ndarray, tuple[int, int]]]:
    """
    Split image into N patches for processing.

    :param image: 2D image array (np.ndarray)
    :param image_patch_dim: Dimension of the image patches (tuple[int, int])
    :param split_image_threshold:
        At which scale relative to the image_patch_dim should images be split (tuple[float, float]).
        For (1.5, 1.5) images are split if their height and width are 1.5X larger than `image_patch_dim`.

    :returns:
        Iterator over image patches as numpy arrays.
    """

    image_height, image_width = image.shape
    target_image_height, target_image_width = image_patch_dim
    scale = (
        image_height / target_image_height,
        image_width / target_image_width,
    )
    if scale <= split_image_threshold:
        yield image, (0, 0)
    else:
        notifications.show_info("Processing image in batches...")
        for i, j in itertools.product(
            range(0, image_height, target_image_height),
            range(0, image_width, target_image_width),
        ):
            yield image[i : i + target_image_height, j : j + target_image_width], (i, j)


def count_objects(
        image: Image,
        roi: Shapes | None,
        model_type: str,
        cellprob_threshold: float,
        flow_threshold: float,
        use_gpu: bool,
        batch_size: int,
        diameter: float
):
    """
    Count objects in an image, or selected ROIs is available.

    :param image: Image layer  (napari.layers.Image)
    :param rois: ROI layer if selected (napari.layers.Shapes | None)
    :param model_type: Segmentation model type (str)
    :param cellprob_threshold: Cell probability threshold (float)
    :param flow_threshold: Flow threshold (float)
    :param restore_type: Image restoration model type (str)
    :param use_gpu: Whether to use GPU acceleration (bool)
    :param diameter: Object diameter (float)

    :returns:
        Segmentation masks (napari.layers.Labels).
        Total number of objects can be accessed with masks.max()
    """

    from cellpose import models
    segmentation_model = models.Cellpose(model_type=model_type, gpu=use_gpu)

    @find_executable_batch_size(starting_batch_size=batch_size)
    def compute_masks(batch_size, model=None):
        count = 0
        if roi is None or len(roi.data) == 0:
            masks0, _, _, _ = model.eval(
                image.data,
                channels=[0,0],
                cellprob_threshold=cellprob_threshold,
                flow_threshold=flow_threshold,
                diameter=diameter,
                batch_size=batch_size
            )
            return masks0, (0, 0), masks0.max()
        else:
            image_roi, roi_offset = get_image_roi(image.data, roi.data[0])
            masks, _, _, _ = model.eval(
                image_roi,
                channels=[0, 0],
                cellprob_threshold=cellprob_threshold,
                flow_threshold=flow_threshold,
                diameter=diameter,
                batch_size=batch_size
            )
            resized_masks = resize(masks, image_roi.shape, order=0, preserve_range=True, anti_aliasing=False).astype(int)
            count = resized_masks.max()

            return masks, roi_offset, count


    masks, mask_offsets, count = compute_masks(model=segmentation_model)
    del segmentation_model

    return masks, mask_offsets, count
